from __future__ import print_function
import six
import hashlib
import os
import time
from random import randint
from binascii import hexlify,unhexlify
from ecdsa import curves, SigningKey
from ecdsa.util import sigencode_der
from .serialize import (
    to_bytes, from_bytes, StreamBaseDecoder, serialize_object, fmt_hex)


__all__ = ('sign_transaction', 'signature_for_transaction')


tfFullyCanonicalSig = 0x80000000

def hash256(data):
    """
        operation twice
    """
    one256 = unhexlify(hashlib.sha256(data).hexdigest())
    return hashlib.sha256(one256).hexdigest()

def get_str(l):
    sss = ""
    while(l>0):
        try:
            l, b = divmod(l, 58)
            sss +=  StreamBaseDecoder.alphabet[b:b+1]
        except Exception:
            print ("get_str error:",six.text_type(b))
            return None
    return sss[::-1]


def sign_transaction(transaction, secret, flag_canonical=True):
    """High-level signing function.hexlify

    - Adds a signature (``TxnSignature``) field to the transaction object.
    - By default will set the ``FullyCanonicalSig`` flag to ``
    """
    if flag_canonical:
        transaction['Flags'] = transaction.get('Flags', 0) | tfFullyCanonicalSig
    sig = signature_for_transaction(transaction, secret)
    transaction['TxnSignature'] = sig
    return transaction


def signature_for_transaction(transaction, secret):
    """Calculate the fully-canonical signature of the transaction.

    Will set the ``SigningPubKey`` as appropriate before signing.

    ``transaction`` is a Python object. The result value is what you
    can insert into as ``TxSignature`` into the transaction structure
    you submit.
    """
    seed = parse_seed(secret)
    key = root_key_from_seed(seed)

    # Apparently the pub key is required to be there.
    transaction['SigningPubKey'] = fmt_hex(ecc_point_to_bytes_compressed(
        key.privkey.public_key.point, pad=True))

    # Convert the transaction to a binary representation
    signing_hash = create_signing_hash(transaction)

    # Create a hex-formatted signature.
    return fmt_hex(ecdsa_sign(key, signing_hash))


def parse_seed(secret):
    """Your Stream secret is a seed from which the true private key can
    be derived.

    The ``Seed.parse_json()`` method of stm-lib supports different
    ways of specifying the seed, including a 32-byte hex value. We just
    support the regular base-encoded secret format given to you by the
    client when creating an account.
    """
    assert secret[0] == 's'
    return StreamBaseDecoder.decode(secret)


def root_key_from_seed(seed):
    """This derives your master key the given seed.

    Implemented in stm-lib as ``Seed.prototype.get_key``
    """
    seq = 0
    while True:
        private_gen = from_bytes(first_half_of_sha512(
            b''.join([seed, to_bytes(seq, 4)])))
        seq += 1
        if curves.SECP256k1.order >= private_gen:
            break

    public_gen = curves.SECP256k1.generator * private_gen

    # Now that we have the private and public generators, we apparently
    # have to calculate a secret from them that can be used as a ECDSA
    # signing key.
    secret = i = 0
    public_gen_compressed = ecc_point_to_bytes_compressed(public_gen)
    while True:
        secret = from_bytes(first_half_of_sha512(
            b"".join([
                public_gen_compressed, to_bytes(0, 4), to_bytes(i, 4)])))
        i += 1
        if curves.SECP256k1.order >= secret:
            break
    secret = (secret + private_gen) % curves.SECP256k1.order

    # The ECDSA signing key object will, given this secret, then expose
    # the actual private and public key we are supposed to work with.
    key = SigningKey.from_secret_exponent(secret, curves.SECP256k1)
    # Attach the generators as supplemental data
    key.private_gen = private_gen
    key.public_gen = public_gen
    return key


def ecdsa_sign(key, signing_hash, **kw):
    """Sign the given data. The key is the secret returned by
    :func:`root_key_from_seed`.

    The data will be a binary coded transaction.
    """
    r, s = key.sign_number(int(signing_hash, 16), **kw)
    r, s = ecdsa_make_canonical(r, s)
    # Encode signature in DER format, as in
    # ``sjcl.ecc.ecdsa.secretKey.prototype.encodeDER``
    der_coded = sigencode_der(r, s, None)
    return der_coded


def ecdsa_make_canonical(r, s):
    """
    Make sure the ECDSA signature is the canonical one.
    """
    # For a canonical signature we want the lower of two possible values for s
    # 0 < s <= n/2
    N = curves.SECP256k1.order
    if not N / 2 >= s:
        s = N - s
    return r, s


def get_stream_from_pubkey(pubkey):
    """
    Given a public key, determine the Stream address.
    """
    ripemd160 = hashlib.new('ripemd160')
    ripemd160.update(hashlib.sha256(pubkey).digest())
    return StreamBaseDecoder.encode(ripemd160.digest())

def get_secret(extra="FSQF5356dsdsqdfEFEQ3fq4q6dq4s5d"):
    """get a random secret"""
    try:
        rnd = hexlify(os.urandom(256))
        tim = time.time()
        data = "%s%s%s%s"%(rnd, tim, randint(100000000000, 1000000000000), extra)
        if six.PY3:
            data = data.encode()
        res = int(hash256(data), 16)
        seed = '21' + str(res)[:32]
        secretKey = hash256(unhexlify(seed))[:8]
        l = int(seed + secretKey, 16)
    except Exception as e:
        print ("get_secret error:",six.text_type(e))
        return None

    return get_str(l)

def get_stream_from_secret(seed):
    """Another helper. Returns the first stream address from the secret."""
    key = root_key_from_seed(parse_seed(seed))
    pubkey = ecc_point_to_bytes_compressed(key.privkey.public_key.point, pad=True)
    return get_stream_from_pubkey(pubkey)


# From stm-lib:hashprefixes.js
HASH_TX_ID = 0x54584E00  # 'TXN'
HASH_TX_SIGN = 0x53545800  # 'STX'
HASH_TX_SIGN_TESTNET = 0x73747800 # 'stx'

def create_signing_hash(transaction, testnet=False):
    """This is the actual value to be signed.

    It consists of a prefix and the binary representation of the
    transaction.
    """
    prefix = HASH_TX_SIGN_TESTNET if testnet else HASH_TX_SIGN
    return hash_transaction(transaction, prefix)


def hash_transaction(transaction, prefix):
    """Create a hash of the transaction and the prefix.
    """
    binary = first_half_of_sha512(
        to_bytes(prefix, 4) +
        serialize_object(transaction, hex=False))
    return hexlify(binary).upper()


def first_half_of_sha512(*bytes):
    """As per spec, this is the hashing function used."""
    hash = hashlib.sha512()
    for part in bytes:
        hash.update(part)
    return hash.digest()[:256//8]


def ecc_point_to_bytes_compressed(point, pad=False):
    """
    In stm-lib, implemented as a prototype extension
    ``sjcl.ecc.point.prototype.toBytesCompressed`` in ``sjcl-custom``.

    Also implemented as ``KeyPair.prototype._pub_bits``, though in
    that case it explicitly first pads the point to the bit length of
    the curve prime order value.
    """

    header = b'\x02' if point.y() % 2 == 0 else b'\x03'
    bytes = to_bytes(
        point.x(),
        curves.SECP256k1.order.bit_length()//8 if pad else None)
    return b"".join([header, bytes])




from binascii import hexlify
from io import BytesIO
from decimal import Decimal
from hashlib import sha256
import types
import six
from six.moves import filter
from six.moves import map
from six.moves import zip
import binascii


__all__ = ('serialize_object',)


##############################################################################
# The following are copied close to verbatim from the stm-lib JavaScript
# source code. We possible could define the types in a simpler way in Python,
# but I want to make it easy to update these structures as the JS lib changes.


# From stm-lib:binformat.js:exports.ledger
LEDGER_ENTRY_TYPES = {
    'AccountRoot': [97],
    'Contract': [99],
    'DirectoryNode': [100],
    'Features': [102],
    'GeneratorMap': [103],
    'LedgerHashes': [104],
    'Nickname': [110],
    'Offer': [111],
    'VStreamState': [114],
    'FeeSettings': [115]
}


# From stm-lib:binformat.js:exports.ter
TRANSACTION_RESULT_VALUES = {
    'tesSUCCESS': 0,
    'tecCLAIM': 100,
    'tecPATH_PARTIAL': 101,
    'tecUNFUNDED_ADD': 102,
    'tecUNFUNDED_OFFER': 103,
    'tecUNFUNDED_PAYMENT': 104,
    'tecFAILED_PROCESSING': 105,
    'tecDIR_FULL': 121,
    'tecINSUF_RESERVE_LINE': 122,
    'tecINSUF_RESERVE_OFFER': 123,
    'tecNO_DST': 124,
    'tecNO_DST_INSUF_STM': 125,
    'tecNO_LINE_INSUF_RESERVE': 126,
    'tecNO_LINE_REDUNDANT': 127,
    'tecPATH_DRY': 128,
    'tecUNFUNDED': 129,
    'tecMASTER_DISABLED': 130,
    'tecNO_REGULAR_KEY': 131,
    'tecOWNERS': 132
}


# From stm-lib:binformat.js:exports.tx, manually constructed
TRANSACTION_TYPES = {
    'AccountSet': 3,
    'TrustSet': 20,
    'OfferCreate': 7,
    'OfferCancel': 8,
    'SetRegularKey': 5,
    'Payment': 0,
    'Contract': 9,
    'RemoveContract': 10,
    'EnableFeature': 100,
    'SetFee': 101
}


# From stm-lib:serializedtypes.js
# Defines the types the binary format supports and the bits
# representing those types.
TYPES_MAP = [
    None,

    # Common:
    'STInt16',    # 1
    'STInt32',    # 2
    'STInt64',    # 3
    'STHash128',  # 4
    'STHash256',  # 5
    'STAmount',   # 6
    'STVL',       # 7
    'STAccount',  # 8

    # 9-13 reserved
    None,    # 9
    None,    # 10
    None,    # 11
    None,    # 12
    None,    # 13

    'STObject',   # 14
    'STArray',    # 15

    # 'Uncommon':
    'STInt8',     # 16
    'STHash160',  # 17
    'STPathSet',  # 18
    'STVector256' # 19
]


# From stm-lib:serializedtypes.js
FIELDS_MAP = {
    # Common types
    1: {# Int16
        1: 'LedgerEntryType',
        2: 'TransactionType'
    },
    2: {# Int32
        2: 'Flags', 3: 'SourceTag', 4: 'Sequence', 5: 'PreviousTxnLgrSeq',
        6: 'LedgerSequence',
        7: 'CloseTime', 8: 'ParentCloseTime', 9: 'SigningTime',
        10: 'Expiration', 11: 'TransferRate',
        12: 'WalletSize', 13: 'OwnerCount', 14: 'DestinationTag',
        # Skip 15
        16: 'HighQualityIn', 17: 'HighQualityOut', 18: 'LowQualityIn',
        19: 'LowQualityOut',
        20: 'QualityIn', 21: 'QualityOut', 22: 'StampEscrow', 23: 'BondAmount',
        24: 'LoadFee',
        25: 'OfferSequence', 26: 'FirstLedgerSequence',
        27: 'LastLedgerSequence', 28: 'TransactionIndex',
        29: 'OperationLimit', 30: 'ReferenceFeeUnits', 31: 'ReserveBase',
        32: 'ReserveIncrement',
        33: 'SetFlag', 34: 'ClearFlag',
    },
    3: {#  Int64
        1: 'IndexNext', 2: 'IndexPrevious', 3: 'BookNode', 4: 'OwnerNode',
        5: 'BaseFee', 6: 'ExchangeRate', 7: 'LowNode', 8: 'HighNode'
    },
    4: {# Hash128
        1: 'EmailHash'
    },
    5: {# Hash256
        1: 'LedgerHash', 2: 'ParentHash', 3: 'TransactionHash',
        4: 'AccountHash', 5: 'PreviousTxnID',
        6: 'LedgerIndex', 7: 'WalletLocator', 8: 'RootIndex',
        16: 'BookDirectory', 17: 'InvoiceID',
        18: 'Nickname', 19: 'Feature'
    },
    6: {# Amount
        1: 'Amount', 2: 'Balance', 3: 'LimitAmount', 4: 'TakerPays',
        5: 'TakerGets', 6: 'LowLimit',
        7: 'HighLimit', 8: 'Fee', 9: 'SendMax', 16: 'MinimumOffer',
        17: 'StreamEscrow'
    },
    7: {# VL
        1: 'PublicKey', 2: 'MessageKey', 3: 'SigningPubKey', 4: 'TxnSignature',
        5: 'Generator',
        6: 'Signature', 7: 'Domain', 8: 'FundCode', 9: 'RemoveCode',
        10: 'ExpireCode', 11: 'CreateCode',12: 'MemoType',13: 'MemoData'
    },
    8: {# Account
        1: 'Account', 2: 'Owner', 3: 'Destination', 4: 'Issuer', 7: 'Target',
        8: 'RegularKey'
    },
    14: {# Object
         1: None,  # end of Object
         2: 'TransactionMetaData', 3: 'CreatedNode', 4: 'DeletedNode',
         5: 'ModifiedNode',
         6: 'PreviousFields', 7: 'FinalFields', 8: 'NewFields',
         9: 'TemplateEntry',10: 'Memo'
    },
    15: {# Array
         1: None, # end of Array
         2: 'SigningAccounts', 3: 'TxnSignatures', 4: 'Signatures',
         5: 'Template',
         6: 'Necessary', 7: 'Sufficient', 8: 'AffectedNodes', 9: 'Memos'
    },

    # Uncommon types
    16: {# Int8
         1: 'CloseResolution', 2: 'TemplateEntryType', 3: 'TransactionResult'
    },
    17: {# Hash160
         1: 'TakerPaysCurrency', 2: 'TakerPaysIssuer', 3: 'TakerGetsCurrency',
         4: 'TakerGetsIssuer'
    },
    18: {# PathSet
         1: 'Paths'
    },
    19: {# Vector256
         1: 'Indexes', 2: 'Hashes', 3: 'Features'
    }
}


INVERSE_FIELDS_MAP = {
    field_name : [type_bit, field_bit]
    for type_bit, fields in FIELDS_MAP.items()
    for field_bit, field_name in fields.items()
}


def serialize_object(obj, hex=True):
    """This is your main entry point to serialize something."""
    stream = BytesIO()
    TypeSerializers.STObject(stream, obj, no_marker=True)
    stream.seek(0)
    bytes = stream.getvalue()
    if hex:
        return fmt_hex(bytes)
    return bytes


def serialize_field(stream, name, value):
    """Binary encode field ``name`` with ``value``, append to ``stream``.
    """
    type_bits, field_bits = INVERSE_FIELDS_MAP[name]
    tag_byte = (type_bits << 4 if type_bits < 16 else 0) | \
               (field_bits if field_bits < 16 else 0)

    if name == 'LedgerEntryType' and isinstance(value, six.string_types):
        value = LEDGER_ENTRY_TYPES[value][0]

    if name == 'TransactionType' and isinstance(value, six.string_types):
        value = TRANSACTION_TYPES[value]

    if name == 'TransactionResult' and isinstance(value, six.string_types):
        value = TRANSACTION_RESULT_VALUES[value]

    if hasattr(value, '__json__'):
        # This indicates it's from our datastructures module, and we
        # need to save the raw value it represents (ex: Amount object).
        value = value.__json__()

    TypeSerializers.STInt8(stream, tag_byte)
    if type_bits >= 16:
        TypeSerializers.STInt8(stream, type_bits)
    if field_bits >= 16:
        TypeSerializers.STInt8(stream, field_bits)

    type_name = TYPES_MAP[type_bits]
    getattr(TypeSerializers, type_name)(stream, value)


def serialize_hex(stream, hexstring):
    """Serialize a hex-encoded value, i.e. '2AE75B908F0'.

    In stm-lib, this is ``serializedtypes.js:serialize_hex()``.
    """
    serialize_bytes(stream, decode_hex(hexstring))


def serialize_bytes(stream, bytes):
    """Serialize a variable length bytestring."""
    serialize_varint(stream, len(bytes))
    stream.write(bytes)


def UInt160(value):
    # In stm-lib, UInt160 is an address. We simply use a string.
    # This helper is the equivalent of UInt160.to_bytes().
    bytes = StreamBaseDecoder.decode(value, 25) # = 20 bytes w/o header??
    return bytes


def serialize_varint(stream, val):
    """
    In stm-lib, this is ``serializedtypes.js:serialize_varint()``.

    Also described here:
        https://stream.com/wiki/Binary_Format#Variable_Length_Data_Encoding
    """
    def rshift(val, n):
        # http://stackoverflow.com/a/5833119/15677
        return (val % 0x100000000) >> n

    assert val >= 0

    bytes = bytearray()
    if val < 192:
        bytes.append(val)
    elif val <= 12480:
        val -= 193
        bytes.append([193 + rshift(val,  8), val & 0xff])
    elif val <= 918744:
        val -= 12481
        bytes.append([
            241 + rshift(val, 16),
            rshift(val, 8) & 0xff,
            val & 0xff
        ])
    else:
        raise ValueError('Variable integer overflow.')

    stream.write(bytes)


class AllStatic(type):
    def __new__(cls, name, bases, attrs):
        for key, value in attrs.items():
            if isinstance(value, types.FunctionType):
                attrs[key] = staticmethod(value)
        return type.__new__(cls, name, bases, attrs)


class TypeSerializers:
    # See stm-lib:serializedtypes.js for the originals.

    __metaclass__ = AllStatic
    def byte_writer(num_bytes):
        def func(stream, value):
            stream.write(to_bytes(int(value), num_bytes))
        return func

    STInt8 = byte_writer(1)
    STInt16 = byte_writer(2)
    STInt32 = byte_writer(4)

    def STAccount(stream, value):
        serialize_bytes(stream, UInt160(value))

    def STAmount(stream, amount):
        if isinstance(amount, dict):
            # non-STM
            hi = 0

            # First bit: non-native
            hi |= 1 << 31

            negative, value, offset = parse_non_native_amount(amount['value'])
            if not value == 0:
                # Second bit: non-negative?
                if not negative:
                    hi |= 1 << 30
                # Next eight bits: offset/exponent
                hi |= ((97 + offset) & 0xff) << 22

            # Remaining 52 bits: mantissa
            # Merge the manually constructed high-bits into the value
            value = value | (hi<<32)

            # Write Amount
            stream.write(to_bytes(value, 8))
            # Write Currency
            TypeSerializers.STCurrency(stream, amount['currency'])
            # Write Issuer
            stream.write(
                StreamBaseDecoder.decode(amount['issuer'], 25))
        else:
            # STM - only support int notation for now, not floats.
            amount = int(amount)

            # Make a 64 bit hex string
            amount_hex = '%x' % abs(amount)
            assert len(amount_hex) <= 16
            amount_hex = amount_hex.zfill(16)

            amount_bytes = bytearray(decode_hex(amount_hex))

            # Clear first bit to indicate STM
            amount_bytes[0] &= 0x3f
            # Clear second bit to indicate negative
            if amount >= 0:
                amount_bytes[0] |= 0x40

            stream.write(amount_bytes)

    def STCurrency(stream, value):
        value = value.upper()
        assert len(value) == 3 and value.isalnum()

        if value == 'STM':
            # Often (like for an Amount object) the currency is not written
            # at all if it is STM. But for example when serializing a path,
            # 'STM' as a currency needs to be indicated, and this is done
            # by 20 empty bytes.
            # Note: As currencies get more complex, demurrage, we will need
            # to introduce a special class; for now we use a simple string.
            # stm-lib:currency.js will help to look at
            stream.write(bytearray(20))

        else:
            data = bytearray(20)
            data[12:15] = map(ord, value)
            stream.write(data)

    def STPathSet(stream, value):
        typeBoundary =  0xff
        typeEnd = 0x00
        typeAccount = 0x01
        typeCurrency = 0x10
        typeIssuer = 0x20
        # A list of paths, mainly used when sending a payment.
        for idx, path in enumerate(value):
            if not idx == 0:
                TypeSerializers.STInt8(stream, typeBoundary)

            for entry in path:
                # For now, the path sets we work with come from
                # streamd directly, with the "type" already set.
                # Still, calculate it new in order to validate.
                type = 0
                if ('account' in entry):
                    type |= typeAccount
                if ('currency' in entry):
                    type |= typeCurrency
                if ('issuer' in entry):
                    type |= typeIssuer
                assert type == entry['type']

                TypeSerializers.STInt8(stream, entry['type'])
                if entry['type'] & typeAccount:
                    stream.write(UInt160(entry['account']))
                if entry['type'] & typeCurrency:
                    TypeSerializers.STCurrency(stream, entry['currency'])
                if entry['type'] & typeIssuer:
                    stream.write(UInt160(entry['issuer']))

        TypeSerializers.STInt8(stream, typeEnd)

    def STVL(stream, value):
        # A variable length string, hex-encoded
        serialize_hex(stream, value)

    def STObject(stream, value, no_marker=False):
        # Ignore lower case field names - non-serializable by convention
        keys = filter(lambda k: not k.islower(), value.keys())
        # Keys need to be sorted
        keys = sort_fields(list(keys))

        for key in keys:
            serialize_field(stream, key, value[key])
        if not no_marker:
            TypeSerializers.STInt8(stream, 0xe1)  # Object ending marker

    def STArray(stream, value):
        for idx, array in enumerate(value):
            keys =  array.keys()
            if (len(keys) != 1):
                raise ValueError("Cannot serialize an array containing non-single-key objects")
            field_name = list(keys)[0];
            val = value[idx][field_name];
            serialize_field(stream, field_name, val)
        TypeSerializers.STInt8(stream, 0xf1);

def sort_fields(keys):
    def sort_key(a):
        type_bits, field_bits = INVERSE_FIELDS_MAP[a]
        return type_bits, field_bits
    keys = keys[:]
    keys.sort(key=sort_key)
    return keys


def parse_non_native_amount(string):
    """Like ``Amount.parse_human()`` in stm-lib, will parse the
    given value into an integer and exponent offset.
    """
    amount = Decimal(string)
    # This will remove trailing zeros
    amount = amount.normalize()

    parts = amount.as_tuple()
    offset = parts.exponent

    # Shift everything such that '10' because '1000000....'
    value = ''.join(map(str, parts.digits))
    full_value = value.ljust(len('9999999999999999'), '0')
    offset -= len(full_value) - len(value)
    full_value = int(full_value)

    # This is special cased
    if full_value == 0:
        offset = -100

    return parts.sign, full_value, offset


def to_bytes(number, length=None, endianess='big'):
    """Will take an integer and serialize it to a string of bytes.

    Python 3 has this, this is originally a backport to Python 2, from:
        http://stackoverflow.com/a/16022710/15677

    We use it for Python 3 as well, because Python 3's builtin version
    needs to be given an explicit length, which means our base decoder
    API would have to ask for an explicit length, which just isn't as nice.

    Alternative implementation here:
       https://github.com/nederhoed/python-bitcoinaddress/blob/c3db56f0a2d4b2a069198e2db22b7f607158518c/bitcoinaddress/__init__.py#L26
    """
    h = '%x' % number
    s = ('0'*(len(h) % 2) + h)
    if length:
        if len(s) > length*2:
            raise ValueError('number of large for {} bytes'.format(length))
        s = s.zfill(length*2)
    s = decode_hex(s)
    return s if endianess == 'big' else s[::-1]


def from_bytes(bytes):
    """Reverse of to_bytes()."""
    # binascii works on all versions of Python, the hex encoding does not
    return int(binascii.hexlify(bytes), 16)


def fmt_hex(bytes):
    """Format the bytes as a hex string, return upper-case version.
    """
    # This is a separate function so as to not make the mistake of
    # using the '%X' format string with an ints, which will not
    # guarantee an even-length string.
    #
    # binascii works on all versions of Python, the hex encoding does not.
    hex = binascii.hexlify(bytes)
    hex = hex.decode()  # Returns bytes, which makes no sense to me
    return hex.upper()


def decode_hex(hex_string):
    """Decode a string like "fa4b21" to actual bytes."""
    if six.PY3:
        return bytes.fromhex(hex_string)
    else:
        return hex_string.decode('hex')


class StreamBaseDecoder(object):
    """Decodes Stream's base58 alphabet.

    This is what stm-lib does in ``base.js``.
    """

    alphabet = 'vpshnaf39wBUDNEGHJKLM4PQRST7VWXYZ2bcdeCg65jkm8oFqi1turAxyz'

    @classmethod
    def decode(cls, *a, **kw):
        """Apply base58 decode, verify checksum, return payload.
        """
        decoded = cls.decode_base(*a, **kw)
        assert cls.verify_checksum(decoded)
        payload = decoded[:-4] # remove the checksum
        payload = payload[1:]  # remove first byte, a version number
        return payload

    @classmethod
    def decode_base(cls, encoded, pad_length=None):
        """Decode a base encoded string with the Stream alphabet."""
        n = 0
        base = len(cls.alphabet)
        for char in encoded:
            n = n * base + cls.alphabet.index(char)
        return to_bytes(n, pad_length, 'big')

    @classmethod
    def verify_checksum(cls, bytes):
        """These stream byte sequences have a checksum builtin.
        """
        valid = bytes[-4:] == sha256(sha256(bytes[:-4]).digest()).digest()[:4]
        return valid

    @staticmethod
    def as_ints(bytes):
        return list([ord(c) for c in bytes])

    @classmethod
    def encode(cls, data):
        """Apply base58 encode including version, checksum."""
        version = b'\x00'
        bytes = version + data
        bytes += sha256(sha256(bytes).digest()).digest()[:4]   # checksum
        return cls.encode_base(bytes)

    @classmethod
    def encode_base(cls, data):
        # https://github.com/jgarzik/python-bitcoinlib/blob/master/bitcoin/base58.py
        # Convert big-endian bytes to integer
        n = int(hexlify(data).decode('utf8'), 16)

        # Divide that integer into base58
        res = []
        while n > 0:
            n, r = divmod(n, len(cls.alphabet))
            res.append(cls.alphabet[r])
        res = ''.join(res[::-1])

        # Encode leading zeros as base58 zeros
        czero = 0 if six.PY3 else b'\x00'
        pad = 0
        for c in data:
            if c == czero:
                pad += 1
            else:
                break
        return cls.alphabet[0] * pad + res


def call_encoder(func, *a, **kw):
    """Test/debug helper to make the stream-based encoder API
    more accessible.

    Equivalent on the stm-lib JS side:

        function encoder(what, value) {
            var SerializedObject = require('../src/js/stream/serializedobject').SerializedObject;
            s = require('../src/js/stream/serializedtypes')
            b = new SerializedObject()
            s[what].serialize(b, value)
            return b.to_hex()
        }
        encoder('Account', 'v3kmLJN5D28dHuH8vZNUZpMC43pEHpaocV')
    """
    def call_util(*a, **kw):
        buffer = BytesIO()
        func(buffer, *a, **kw)
        buffer.seek(0)
        return fmt_hex(buffer.getvalue())
    if a or kw:
        return call_util(*a, **kw)
    return call_util



from decimal import Decimal
import six
import json
if six.PY2:
    import __builtin__ as builtins
else:
    import builtins
from six.moves import filter


# During debugging, this helps with the StreamPrimitive __getattr__
# problem, described further down below.
#class property(property):
#    def __get__(self, obj, cls):
#        try:
#            return super(property, self).__get__(obj, cls)
#        except AttributeError, e:
#            import traceback
#            traceback.print_exc()
#            import sys
#            sys.exit(1)


class StreamPrimitive(dict):
    """Dict that allows attribute access."""

    def __init__(self, data):
        dict.__init__(self, **data)

    def __getattr__(self, item):
        try:
            value = self[item]
            # Convert nested dicts into StreamPrimitives on the fly on access
            if isinstance(value, dict) and not isinstance(value, StreamPrimitive):
                value = StreamPrimitive(value)
                self[item] = value
            return value
        except KeyError:
            # XXX: This keeps hiding real AttributeErrors that occur during
            # handling of a @property; if property "foo" somehow triggers
            # an AttributeError, this will be called and will expose the
            # exception we raise here instead of the original one. See also
            # discussed here amongst other places:
            # http://stackoverflow.com/questions/15401180/using-getattr-and-meeting-expected-behaviour-for-subclasses
            # (see also the related Python bug report).
            # One possible way to fix this might be implementing
            # __getattribute__ instead and check if a property exists.
            # Until then, enable the hack on top of this file.
            raise AttributeError(item)

    def __repr__(self):
        return '<%s %s>' % (self.__class__.__name__, dict.__repr__(self))

    def __unicode__(self):
        return json.dumps(self.__json__())

    def __json__(self):
        return dict.copy(self)

    def copy(self):
        return StreamPrimitive(self)


class tupledict(list):
    """A list of 2-tuples that can be used as a dict."""
    def __getitem__(self, item):
        for key, value in self:
            if key == item:
                return value
        return list.__getitem__(self, item)


class VStreamStateEntry(StreamPrimitive):
    """Stream state entries exist when one account sets a credit limit
    to another account in a particular currency or if an account holds
    the IOUs of another account. Each entry is shared between two accounts.
    """

    def affects_account(self, account):
        try:
            if self.counter_party(account):
                return True
        except ValueError:
            return False

    def counter_party(self, account):
        if self.LowLimit.issuer == account:
            return self.HighLimit.issuer
        if self.HighLimit.issuer == account:
            return self.LowLimit.issuer
        raise ValueError('%s is not a party' % account)

    def balance(self, account):
        """Return the balance from the perspective of the given account.
        """
        if self.LowLimit.issuer == account:
            return Decimal(self.Balance.value)
        if self.HighLimit.issuer == account:
            return -Decimal(self.Balance.value)
        raise ValueError('%s is not a party' % account)

    def trust_limit(self, account):
        """Return the trust limit from the perspective of the given account.
        """
        if self.LowLimit.issuer == account:
            return Decimal(self.LowLimit.value)
        if self.HighLimit.issuer == account:
            return Decimal(self.HighLimit.value)
        raise ValueError('%s is not a party' % account)


class Amount(object):
    """Represents a Stream amount. The datastructure behind this will
    either be a dict with amount/currency/issuer keys, or a natural
    number, in which case we are dealing with STM.

    Usage/Features:

    - The properties are used to normalize access.
    - Supports arithmetic.
    - Can be initialized with with special formats:
        Decimal('1') for STM for drops, for example.

    Note: This ideally should inherit StreamPrimitive, but then we have
    trouble serializing it to an integer in the STM case, since
    StreamPrimitive as a dict, and Python's json module does not provide
    any hook to change how basic types are handled.
    """

    def __init__(self, data):
        # In Stream data structures, data is either a IOU dict, or
        # STM drops in int. We want to allow the developer to init
        # an Amount object with other values.
        if isinstance(data, six.string_types):
            # Treat as STM, convert to drops, if there is a decimal
            # point. Otherwise we have to support this as drops, since
            # streamd often returns drops as string objects.
            if '.' in data:
                data = Decimal(data)
                # Fall-through

        if isinstance(data, Decimal):
            # If a decimal is given,
            data = Decimal(data) * stm_base
            assert int(data) == data
            data = int(data)

        self.data = data

    @property
    def currency(self):
        if isinstance(self.data, dict):
            return self.data['currency']
        else:
            return 'STM'

    @property
    def issuer(self):
        if isinstance(self.data, dict):
            return self.data['issuer']
        else:
            return None

    def _get_value(self):
        if isinstance(self.data, dict):
            return Decimal(self.data['value'])
        else:
            return stm(self.data)
    def _set_value(self, v):
        if isinstance(self.data, dict):
            self.data['value'] = '%s' % v
        else:
            self.data = int(Decimal(v) * stm_base)
    value = property(_get_value, _set_value)

    def copy(self, new_value=None):
        copy = Amount(self.data.copy() if isinstance(self.data, dict) else self.data)
        if new_value is not None:
            copy.value = new_value
        return copy

    def __unicode__(self):
        return '%s' % self.value

    def __json__(self):
        return self.data

    def __contains__(self, item):
        return item in self.data

    def __getitem__(self, item):
        return self.data[item]

    def __setitem__(self, key, value):
        self.data[key] = value

    def _assert_compat_other(self, other):
        """For arithmetic with the Amount class, check that the ``other``
        object can be handled.
        """
        if isinstance(other, str):
            return Decimal(other)
        if isinstance(other, Decimal):
            return other
        assert other.currency == self.currency
        return other.value

    def __add__(self, other):
        other_value = self._assert_compat_other(other)
        new_value = self.value + other_value
        return self.copy(new_value)

    def __sub__(self, other):
        other_value = self._assert_compat_other(other)
        new_value = self.value - other_value
        return self.copy(new_value)

    def __div__(self, other):
        other_value = self._assert_compat_other(other)
        new_value = self.value / other_value
        return self.copy(new_value)

    def __rdiv__(self, other):
        return Amount.__truediv__(other, self)



class AccountRootEntry(StreamPrimitive):
    """An account root entry exists for each account. It holds its STM
    balance, last transaction sequence number, and related information.
    """

    def affects_account(self, account):
        if not 'Account' in self:
            # A transactions AffectedNodes can contain an AccountRoot
            # node entry without fields, so account for this.
            return False
        return self.Account == account


class OfferEntry(StreamPrimitive):
    """A offer entry specifies the terms of exchange between two currencies.
    """

    def affects_account(self, account):
        # Does not return True if the account's IOUs are being traded
        return self.Account == account


class DirectoryNodeEntry(StreamPrimitive):

    def affects_account(self, account):
        return False


LedgerEntries = {
    'AccountRoot': AccountRootEntry,
    'VStreamState': VStreamStateEntry,
    'Offer': OfferEntry,
    'DirectoryNode': DirectoryNodeEntry,
}


def shadow(front, back):
    # For now this is a hard-merge, but I'd prefer a transient fall-through.
    result = back.copy()
    result.update(front)
    return result


class NodeModification(StreamPrimitive):
    """An entry in the ``AffectedNodes`` key of a processed transaction.
    """

    def __init__(self, data):
        StreamPrimitive.__init__(self, data)
        node_class = LedgerEntries[data['LedgerEntryType']]
        self.new = node_class(data.get('FinalFields', {}))
        if 'PreviousFields' in data:
            self.old = node_class(
                # PreviousFields only contains parts, presumably those
                # that changed, so add a fallback to the shared data).
                shadow(data['PreviousFields'], data.get('FinalFields', {})))
        else:
            self.old = None
        self.type = type(self.new)

    def __getattr__(self, item):
        return getattr(self.new, item)


class NodeCreation(NodeModification):
    """An entry in the ``AffectedNodes`` key of a processed transaction.
    """

    def __init__(self, data):
        StreamPrimitive.__init__(self, data)
        node_class = LedgerEntries[data['LedgerEntryType']]
        self.new = node_class(data['NewFields'])
        self.old = None
        self.type = type(self.new)


class NodeDeletion(NodeModification):
    """An entry in the ``AffectedNodes`` key of a processed transaction.
    """


class Transaction(StreamPrimitive):
    """Makes data from a stream transaction structure accessible.

    ``meta`` can be given if the transaction data has no ``metaData``
    key. This is because the Stream server will hand out different
    formats: When querying a ledger, the meta that is in said key.
    When subscribing to the transaction feed, the metadata is given
    separately from the transaction.
    """

    def __init__(self, data, meta=None):
        StreamPrimitive.__init__(self, data)
        self.meta = meta

        # __new__ could also be used for this I suppose
        subclass = {
            'Payment': PaymentTransaction,
            'OfferCreate': OfferCreateTransaction,
            'OfferCancel': OfferCancelTransaction,
            'TrustSet': TrustSetTransaction,
            'AccountSet': AccountSetTransaction,
            'SetRegularKey': SetRegularKeyTransaction,
            'EnableAmendment': EnableAmendmentTransaction,
            'SeeFee': SetFeeTransaction,
            'SignerListSet': SignerListSetTransaction,
            'PaymentChannelFund': PaymentChannelFundTransaction,
            'PaymentChannelCreate': PaymentChannelCreateTransaction,
            'PaymentChannelClaim': PaymentChannelClaimTransaction,
            'EscrowFinish': EscrowFinishTransaction,
            'EscrowCreate': EscrowCreateTransaction,
            'EscrowCancel': EscrowCancelTransaction,
        }.get(data['TransactionType'], UnknownTransaction)
        self.__class__ = subclass

    @property
    def type(self):
        return type(self)

    def _get_meta(self):
        return self._meta or self.metaData
    def _set_meta(self, value):
        self._meta = StreamPrimitive(value) if value else value
    meta = property(_get_meta, _set_meta)

    @property
    def successful(self):
        return self.meta.TransactionResult == 'tesSUCCESS'

    @property
    def affected_nodes(self):
        for node in self.meta.AffectedNodes:
            assert len(list(node.keys())) == 1
            change_type = list(node.keys())[0]
            node_class = {
                'CreatedNode': NodeCreation,
                'ModifiedNode': NodeModification,
                'DeletedNode': NodeDeletion}[change_type]
            yield node_class(list(node.values())[0])

    def _get_nodes(self, account=None, type=None):
        """Return affected nodes matching the filters."""
        result = self.affected_nodes
        if account:
            # Allow to filter by multiple accounts
            if not isinstance(account, list):
                account = [account]
            result = filter(
                lambda n: all([n.affects_account(a) for a in account]), result)
        if type:
            result = filter(lambda n: n.type == type, result)
        return list(result)

    def _get_node(self, account=None, type=None):
        """Return a affected node matching the filters, and make sure
        there is only one."""
        result = self._get_nodes(account, type)
        assert len(result) == 1, 'One node expected, found %s' % len(result)
        return result[0]


stm_base = Decimal('1000000')

def stm(s):
    """STM is given in the API as a large int, convert to a decimal.
    """
    return Decimal(s) / stm_base

def stm_drops(s):
    """convert STM to large int.
    """
    return (Decimal(s) * stm_base).quantize(Decimal('1'))


class first(object):
    """Provide a simplified accessor for a property that returns
    multiple values.

    Assumes the property value has the following format::

        [(key, value), (key, value)]

    Will return prop[0].value if there is a single item, or raise an error.
    """
    def __init__(self, attr):
        self.attr = attr
    def __get__(self, instance, owner):
        multiple = getattr(instance, self.attr)
        if len(multiple) > 1:
            raise ValueError('More than one issuer on recipient side, '
                             'use the multi-value access property')
        elif len(multiple) == 0:
            return None
        else:
            return multiple[0][1]


class PaymentTransaction(Transaction):

    @property
    def num_received_issuers(self):
        """The number of different issuers received.

        Returns 0 in case of STM.
        """
        if self.is_stm_received:
            return 0
        return len(self.currencies_received[1])

    @property
    def currencies_received(self):
        """Returns a 2-tuple (code, issuer) indicating the currency
        that was received. In case of STM, ``('STM', None)`` is returned.

        What it does:

        The currency itself is readily available in ``Amount.currency``
        (or in case of STM, ``Amount`` will be an integer). The issuer
        of that currency is a bit more complicated. Here are some places
        where we do not find it:

        - ``Amount.issuer`` - always seems to be the account of the recipient.
        - The last element of ``Paths``. Frequently, multiple paths are
          listed, and it's not clear which one the transaction took.

        All in all, it it seems as if that part of the transaction is verbatim
        what the client submitted. Instead, we look into ``metaData``.

        There, we find a list of AffectedNodes, and we just find the one
        that relates to the account of the recipient.
        """
        if self.is_stm_received:
            # This means STM was received.
            return ('STM', None)
        else:
            return (
                self.Amount.currency,
                [node.counter_party(self.Destination) for node in self._get_nodes(
                    account=self.Destination, type=VStreamStateEntry)]
            )

    @property
    def amounts_received(self):
        """A list of all the amounts received by issuer.

        If there is only one issuer, the output is similar to what you'd
        see from :prop:`amount_received`.
        """
        result = []
        for node in self._get_nodes(
                account=self.Destination, type=VStreamStateEntry):
            result.append((
                node.new.balance(self.Destination)
                    - node.old.balance(self.Destination),
                self.Amount.currency,
                node.counter_party(self.Destination)))
        return result

    @property
    def amount_received(self):
        """3-tuple of (amount, currency, issuers), representing the full
        amount received.
        """
        if self.meta.get('DeliveredAmount'):
            amount = self.meta.DeliveredAmount.value \
                if isinstance(self.meta.DeliveredAmount, dict) \
                else stm(self.meta.DeliveredAmount)
        else:
            amount = self.Amount.value \
                if isinstance(self.Amount, dict) else stm(self.Amount)
        return tuple(
            [amount] + list(self.currencies_received)
        )

    @property
    def is_stm_received(self):
        # Looking at this field is the only way. It should be reliable.
        # In complex routings, in don't thing AffectedNodes can conclusively
        # tell us; or, it's really hard.
        return not isinstance(self.Amount, dict)

    @property
    def is_stm_sent(self):
        return not isinstance(self.SendMax, dict)

    def get_balances(self, who, previous=False):
        """Returns the previous balances with each issuer.
        """
        where = 'old' if previous else 'new'
        if self.is_stm_received:
            # If it is a STM payment, there should be one AccountRoot change
            node = self._get_node(account=who, type=AccountRootEntry)
            return [(None, stm(getattr(node, where).Balance))]
        else:
            # Otherwise, there should be one or more VStreamState entries
            # for each issuer.
            nodes = self._get_nodes(account=who, type=VStreamStateEntry)
            return tupledict(
                [(node.counter_party(self.Destination),
                 getattr(node, where).balance(who)) for node in nodes])

    @property
    def recipient_balances(self):
        return self.get_balances(self.Destination)

    @property
    def recipient_previous_balances(self):
        return self.get_balances(self.Destination, previous=True)

    @property
    def recipient_trust_limits(self):
        if self.is_stm_received:
            return []
        else:
            nodes = self._get_nodes(account=self.Destination, type=VStreamStateEntry)
            return tupledict(
                [(node.counter_party(self.Destination),
                 node.new.trust_limit(self.Destination)) for node in nodes])

    @property
    def sender_trust_limits(self):
        """The trust limits of the sender changed in this transaction.

        Multiple of a sender's trust limits may have changed during the
        transaction, because the full amount of currency sent may be
        a combination of different balances.

        """
        nodes = self._get_nodes(account=self.Account, type=VStreamStateEntry)
        return tupledict(
            [(node.counter_party(self.Account),
             node.new.trust_limit(self.Account)) for node in nodes])

    # If there is only one issuer being received, make access easier.
    # These raise an exception when more than one issuer is involved.
    recipient_balance = first('recipient_balances')
    recipient_previous_balance = first('recipient_previous_balances')
    recipient_trust_limit = first('recipient_trust_limits')
    sender_trust_limit = first('sender_trust_limits')

    def analyze_path(self):
        """This will give you some information about how the payment was
        routed.

        Specifically, the return value is a dict that looks like this::

            {'intermediaries': 2, 'offers': 1}

        If intermediaries is 0, there was a trust set between the two parties
        (or its a direct STM payment).
        If intermediaries is 1, both parties trust the same third party.
        If intermediaries is 2, one additional user helped.
        And so on.

        ``offers`` specifies the number of market offers that were fully
        or partially executed during payment routing. This is always included
        in ``intermediaries``.

        So the following may be true:

            intermediaries - offers - 1 = streams involved

        ----

        How is this done? Drawing automated conclusions from the set of
        unlinked node changes can be a bit of a challenge. For example, an
        AccountRoot node may be modified when someone pays in STM, when
        a fee is claimed during an IOU payment, or when your offer gets
        resolved during a third party's payment.
        Presumably, this can be vastly improved.

        TODO: In particular, in cases where the recipient receives currency
        from multiple issuers, there really should be a way to make
        this info more accessible - maybe see the individual paths
        separately.
        """

        # list() used once exhausts the generator, it's tiring..
        filter = lambda f, d: list(builtins.filter(f, d))

        # Ignore all DirectoryNodes, not sure what they do, it seems
        # like upkeep.
        nodes = filter(lambda n: n.type != DirectoryNodeEntry, self.affected_nodes)

        # Ignore all STM acounting nodes. These either indicate a fee,
        # or a direct payment, which would mean no intermediaries.
        nodes = filter(lambda n: n.type != AccountRootEntry, nodes)

        # Ignore all nodes involving the recipient. If its a direct payment,
        # it will delete the sender's state node as well. If we are dealing
        # with a third party IOUs, the sender's balance node will be left to
        # be counted as our "one" hop
        nodes = filter(lambda n: not n.new.affects_account(self.Destination), nodes)

        # Count the offer nodes. These are easy, each such node indicates
        # one offer that was involved.
        offers = filter(lambda n: n.type == OfferEntry, nodes)

        # Each offer comes with AccountRoot and VStreamStateEntry nodes
        # for the accounts of the offerer, so we need to filter those
        # out as well.
        for offer in offers:
            nodes = filter(lambda n: not n.affects_account(offer.Account), nodes)

        # What is left is the payee VStreamState + one VStreamState for each
        # true intermediary that was involved.
        # For validation, make sure all the nodes left are VStreamState entries
        assert len(filter(lambda n: n.type != VStreamStateEntry, nodes)) == 0

        return {
            'intermediaries': len(nodes) + len(offers),
            'offers': len(offers)
        }


class OfferCreateTransaction(Transaction):
    pass


class OfferCancelTransaction(Transaction):
    pass


class TrustSetTransaction(Transaction):
    pass


class AccountSetTransaction(Transaction):
    pass


class SetRegularKeyTransaction(Transaction):
    pass


class EnableAmendmentTransaction(Transaction):
    pass


class SetFeeTransaction(Transaction):
    pass


class SignerListSetTransaction(Transaction):
    pass


class EscrowCreateTransaction(Transaction):
    pass


class EscrowCancelTransaction(Transaction):
    pass


class EscrowFinishTransaction(Transaction):
    pass


class PaymentChannelClaimTransaction(Transaction):
    pass


class PaymentChannelCreateTransaction(Transaction):
    pass


class PaymentChannelFundTransaction(Transaction):
    pass


class UnknownTransaction(Transaction):
    pass



class TransactionSubscriptionMessage(StreamPrimitive):
    """The data structure returned by the server when subscribing to
    transaction updates.
    """

    @property
    def transaction(self):
        return Transaction(self['transaction'], meta=self['meta'])



from __future__ import unicode_literals
import six
if six.PY3:
    from queue import Queue, Empty
    import urllib.parse as parse
else:
    from Queue import Queue, Empty
    import urllib as parse
from decimal import Decimal
import json
import threading
import websocket
import logging
from .datastructures import Amount,stm,stm_drops
from .serialize import serialize_object,fmt_hex
from .sign import hash_transaction, HASH_TX_ID, get_stream_from_secret, sign_transaction

__all__ = ('Remote', 'Client', 'StreamError')


log = logging.getLogger('stream.client')
log.addHandler(logging.NullHandler())


class StreamError(Exception):
    pass


class ResponseError(StreamError):
    def __init__(self, error_response):
        self.response = error_response
        StreamError.__init__(
            self,
            error_response.get('error_exception',
                               error_response.get('error_message', '')))

    def __getitem__(self, item):
        return self.response[item]


class TransactionError(ResponseError):
    """Respresents an error in a transaction. A transaction is a
    higher-level object of our own. For example, the stream server
    may give us a temporary "tes*" response, and we continue to wait
    and confirm until we have a final answer. Only then do we complete
    or fail our transaction.
    """
    def __init__(self, error_code, full_response):
        self.response = full_response
        StreamError.__init__(
            self,
            full_response.get('error_exception',
                               full_response.get('error_message', '')))

    def __getitem__(self, item):
        return self.response[item]


FEE_DEFAULTS = {
    'fee_ref': 10,       # cost of reference transaction in "fee units"
    'fee_base': 10,      # cost of reference transactios in STM
    'load_base': 256,
    'load_factor': 256,
    'base_fee': 10       # default cost of a transaction in units
}


class StreamEncoder(json.JSONEncoder):
    def default(self, obj):
        if hasattr(obj, '__json__'):
            return obj.__json__()
        return json.JSONEncoder.default(self, obj)


def transaction_hash(tx_json):
    """To identify a transaction, we can use the signing-hash algorithm
    with a special prefix; the result should match the transaction id
    that the server will use.

    In particular because we do not have a blob serializer, we'll have to
    use the dict structure to calculate the hash.
    """
    return hash_transaction(tx_json, HASH_TX_ID)


class DeferredResponse(object):
    """A future that can either return a result value or raises
    an exception.

    - Must be resolved with a stream server response, or an exception
      instance.
    - If the response indicates an error, will throw a ResponseError.
    - If resolved with an exception instance, will throw that.
    - Otherwise will resolve the the stream server response.
    """
    def __init__(self):
        self.resolved = threading.Event()
        self.response = None
        self.resulter = None

    def wait(self, timeout=None):
        self.resolved.wait(timeout)

        if isinstance(self.response, Exception):
            raise self.response

        if not self.response['status'] == 'success':
            raise ResponseError(self.response)
        result = self.response['result']
        if self.resulter:
            result = self.resulter(result)
        return result

    def resolve(self, response):
        self.response = response
        self.resolved.set()


class SubscriptionQueue(Queue):

    def get(self, block=True, timeout=None):
        result = Queue.get(self, block, timeout)
        if isinstance(result, Exception):
            raise result
        return result



class Client(object):
    """Connection to Stream network.

    This is supposed to be more of a low-level API, representing
    a single connection. You'll have to handle errors manually.

    It uses a thread internally, and is itself threadsafe. Subscriptions
    return a queue that can be read from.

    Some notes on the design
    ------------------------

    I still haven't grogged how to do network APIs. I think my confusion
    stems from the desire to enable both blocking and non-blocking use
    cases, hating threads, nor wanting to rely on greenlets.

    Here are some things that were considered:

    1. Expose a Deferred object to the user. This feels like re-inventing
       Twisted (up to API methods needing to inject code "on success" to
       fetch the desired data from the result structure).
       This approach means the user has an async-interface within a single
       thread of his own, which seems a waste considering we only have
       a single websocket connection to begin with.

    2. Keep the wait() internal, the external API methods block until a
       result is available. Makes for a nice sync API while still letting
       the user use multiple threads if he so desires (like handling the
       results of a subscribe). This is what we are doing now.

    3. Use no thread thread at all, not allowing async usage. That
       means the user has to clear subscription messages in this main
       thread in between regular interactions. I felt this would be too
       restrictive.

    4. Do not use an internal thread but still support async usage. I
       believe this to be possible along the lines of this pseudo code::

           def consume_one(self):   # read subscription messages
                dispatch_queue_first
                with self.recv_lock:
                    msg = self.conn.recv(block=False)
                    dispatch_queue_again
                    if msg: dispatch_msg(triggers command wait() events)

            def wait(self, id):
                check_queue_first
                with self.recv_lock:
                    check_queue_again
                    receive_and_add_to_queue(until msg with id found)

        The downside: while waiting for one response, all other readers
        are blocked, even if a response for their command has come in.
        Given that responses should generally be immediate, presumably
        this isn't that big a deal. Still, it felt a bit hacky.


    How should subscriptions be handled? In general, when a thread is
    used, we'd either have to do the callback from the thread, which
    sucks, or, what I'm doing know, giving the consumer a queue which
    he'll have to read from at his convenience.
    """
    # TODO: Better handle keyboard interrupts while waiting:
    #    http://stackoverflow.com/a/14421297/15677

    def __init__(self, url):
        self.fee_info = FEE_DEFAULTS.copy()

        # These will be used to sync the reading thread with the threads
        # that are consuming us. Yes, single dict and lock could be used,
        # but when setting up a subscription we need to wait for a server
        # ok but can't miss a message afterwards, so a separate lock
        # allows us to block incoming subscriptions update while still
        # setting up the callback queue.
        self.callbacks = {}
        self.subscriptions = {}
        self.callbacks_lock = threading.RLock()
        self.subscriptions_lock = threading.RLock()

        # TODO: We need to deal with timeouts (a ping thread?)
        self.conn = websocket.create_connection(url, timeout=30)
        # This thread will do the reading in the basis for in turn
        # supporting multiple threads to use *this* class.
        self._read_thread = thread = threading.Thread(target=self._read_proc)
        thread.setDaemon(True)
        thread.start()

    def close(self):
        log.debug('client.close()')
        self._shutdown = True
        self.conn.close()

    def _mkid(self):
        setattr(self, '_id', getattr(self, '_id', 0) + 1)
        return self._id

    def _read_proc(self):
        """Runs the reading thread."""
        try:
            while not getattr(self, '_shutdown', False):
                if six.PY2:
                    msg = json.loads(self.conn.recv().decode('utf-8'))
                else:
                    msg = json.loads(self.conn.recv())
                log.debug('<<<<<<<< receiving % s', json.dumps(msg, indent=2))

                type = msg['type']

                # Response to a regular command
                if type == 'response':
                    with self.callbacks_lock:
                        if msg['id'] in self.callbacks:
                            self.callbacks[msg['id']].resolve(msg)
                            continue

                # Else this will be a subscription response
                with self.subscriptions_lock:
                    if type in self.subscriptions:
                        # Multiple subscriptions may have been issued for
                        # the same type, notify all of them.
                        for queue in self.subscriptions[type]:
                            queue.put(msg)
                        continue

                raise ValueError(
                    'unexpected message from server: %s' % str(msg))
        except Exception as e:
            # If we have already shutdown, ignore the error. This is
            # because by shutting down the socket during a recv(), a
            # variety of socket-related / SSL errors are to be expected.
            if not getattr(self, '_shutdown', False):
                # Notify all callbacks so that exceptions occur
                # in all waiters.
                with self.callbacks_lock:
                    for deferred in self.callbacks.values():
                        deferred.resolve(e)
                with self.subscriptions_lock:
                    for queues in self.subscriptions.values():
                        for queue in queues:
                            queue.put(e)
                # Also shut down the connection so that the main thread
                # doesn't keep sending while not getting a response.
                self.conn.close()
                # Finally, re-raise the exception in the thread
                raise
        log.debug('client.read_proc now shut down')

    def execute(self, cmd, **data):
        """Send a commad to the server, wait for the result. Sync!

        Possible outcomes are from DeferredResponse.wait(): A stream
        server response or an ResponseError exception.
        """
        # Prepare the command to send
        data['command'] = cmd
        data['id'] = self._mkid()
        data = {k:v for k, v in data.items() if v is not None}

        log.debug('>>>>>>>> sending %s', json.dumps(data, indent=2, cls=StreamEncoder))
        if six.PY2:
            self.conn.send(json.dumps(data, cls=StreamEncoder).encode('utf-8'))
        else:
            self.conn.send(json.dumps(data, cls=StreamEncoder))
        with self.callbacks_lock:
            self.callbacks[data['id']] = DeferredResponse()
        msg = self.callbacks[data['id']].wait()
        return msg

    def subscribe(self, streams=None):
        def add_queue(name, queue):
            queues = self.subscriptions.setdefault(name, [])
            queues.append(queue)

        with self.subscriptions_lock:
            result = self.execute('subscribe', streams=streams)
            self._process_fee_update(result)
            # Setup a queue for subscription messages
            queue = SubscriptionQueue()
            for stream in streams:
                if 'ledger' == stream:
                    add_queue('ledgerClosed', queue)
                elif 'transactions' == stream:
                    add_queue('transaction', queue)
                elif 'server' == stream:
                    add_queue('serverStatus', queue)
                else:
                    raise ValueError(stream)

        return result, queue

    def add_fee(self, tx, amount=None, cushion='1.2'):
        """Add a fee to the given transaction dict.
        """
        if not amount:
            # Note: The stream client uses a design where every transaction
            # may have a different cost in fee units, but then just uses
            # 10 as a default. We'll ignore it.
            D = Decimal
            i = lambda k: D(self.fee_info[k])
            # One fee unit in STM
            one_unit = i('fee_base') / i('fee_ref')
            # Therefore the cost =
            fee = one_unit * i('base_fee')
            # Consider the load
            fee = fee * (i('load_factor') / i('load_base'))
            # Add a safety cushion
            amount = int(fee * D(cushion))
        tx['Fee'] = amount

    def _process_fee_update(self, msg):
        """Call this with a message like a server_info response. Will
        affect fee calculations."""
        i = self.fee_info
        # Result to a subscribe command will contain those
        i['load_base'] = msg.get('load_base', i['load_base'])
        i['load_factor'] = msg.get('load_factor', i['load_factor'])

        # Note: the ledger stream gives us data for fee_ref and fee_base.
        # The stream client uses these to calculate the STM value of a
        # fee unit, though the Wiki instructions to calculate fees do not
        # mention it. We'll ignore it for now.

    def request_account_info(self, account):
        return self.execute("account_info", account=account)['account_data']

    def request_account_tx(self, account,ledger_index_min,ledger_index_max):
        if(ledger_index_max < 0):
            return self.execute("account_tx", account=account,ledger_index_min=ledger_index_min)
        else:
            return self.execute("account_tx", account=account,ledger_index_min=ledger_index_min,ledger_index_max=ledger_index_max)

    def request_account_lines(self, account):
        return self.execute("account_lines", account=account)['lines']

    def book_offers(self, taker_gets, taker_pays, limit):
        return self.execute("book_offers", taker_gets=taker_gets, taker_pays=taker_pays, limit=limit)

    def account_offers(self, account):
        return self.execute("account_offers", account=account)

    def tx(self, transaction):
        return self.execute("tx", transaction=transaction)

    def submit(self, tx_blob=None, tx_json=None, secret=None):
        """Submit the transaction.

        You may either send a signed transaction (as a serialized
        tx_blob or unserialized dict, both as ``tx_blob``), or you
        may pass a secret and a ``tx_json`` dict to let the server
        do the signing.
        """
        assert not (tx_blob and tx_json)
        if isinstance(tx_blob, dict):
            tx_blob = serialize_object(tx_blob)
        return self.execute(
            "submit", tx_blob=tx_blob, tx_json=tx_json, secret=secret)

    def find_path_once(self, source, destination, destination_amount,
                       source_currencies=None):
        """Use the ``stream_path_find`` API to find a path"""
        return self.execute(
            "path_find", source_account=source,
            destination_account=destination,
            destination_amount=destination_amount,
            source_currencies=source_currencies
        )

    def path_find(self):
        # Start an ongoing path-finding process. Uses the ``path_find``
        # API instead.
        raise NotImplementedError()


class DeferredTransaction(object):
    """The difference to DeferredResponse is that this one will
    always be resolved with an actual transaction result. However,
    the transaction may indicate a failure.
    """
    # TODO: Can we just re-use DeferredResponse? - at least extend it

    def __init__(self, tx, txhash):
        self.tx = tx
        self.hash = txhash
        self.resolved = threading.Event()
        self.result = self.error = None

    def wait(self, timeout=None):
        self.resolved.wait(timeout)
        if self.error:
            raise TransactionError(self.error, self.result)
        return self.result

    def resolve(self, result, error=None):
        self.result = result
        self.error = error
        self.resolved.set()


class Remote(object):
    """This is supposed to be a more high-level API that ideally
    will be able to manage multiple server connections, can track
    the state of submitted transactions etc.

    This is partially async. That is, it blocks while waiting for
    direct server responses, but is async for transaction handling,
    where the server cannot be expected to give a final response
    without significant delay.
    """

    def __init__(self, url, secret):
        self.secret = secret
        self._sequence_cache = {}
        self._pending_transactions = {}
        self._pending_transactions_lock = threading.RLock()

        # Connect to the client
        self.client = Client(url)
        # Start a subscription to transaction updates
        # TODO: Should we only do this once we begin sending payments?
        _, queue = self.client.subscribe(streams=['server', 'transactions'])

        # This thread will deal with subscription updates
        self.read_thread = threading.Thread(target=self._read_proc, args=(queue,))
        self.read_thread.setDaemon(True)
        self.read_thread.start()

    def _read_proc(self, queue):
        try:
            while not getattr(self, '_shutdown', False):
                try:
                    msg = queue.get(timeout=0.2)
                except Empty:
                    continue
                if msg['type'] == 'serverStatus':
                    # TODO: Thread: needs to lock!
                    self.client._process_fee_update(msg)

                if msg['type'] == 'transaction':
                    # See if this is a transaction that interests us
                    hash = msg['transaction']['hash']
                    with self._pending_transactions_lock:
                        if hash in self._pending_transactions:
                            # The JS client, when a transaction comes in,
                            # doesn't seem to check any field except
                            # validated, and then just considered the
                            # transaction a success, which I find a bit
                            # strange. But do likewise.
                            if not msg['validated']:
                                msg = StreamError(
                                    'received non-validated transaction, is '
                                    'this legit? %s' % msg)

                            self._pending_transactions[hash].resolve(msg)
                            del self._pending_transactions[hash]

        except Exception as e:
            # On error, notify all watches
            with self._pending_transactions_lock:
                for transaction in self._pending_transactions.values():
                    transaction.resolve(e)
            raise

        log.debug('remote.read_proc now shut down')

    def close(self):
        log.debug('remote.close()')
        self._shutdown = True
        self.client.close()

    def get_sequence_number(self, account):
        if not account in self._sequence_cache:
            # Will update the cache
            self.account_info(account)
        current = self._sequence_cache[account]
        self._sequence_cache[account] += 1
        return current

    def account_info(self, account):
        info = self.client.request_account_info(account)
        self._sequence_cache[account] = info['Sequence']
        return info

    def account_tx(self, account,ledger_index_min = -1,ledger_index_max = -1):
        if(ledger_index_max < 0):
            info = self.client.request_account_tx(account,ledger_index_min,-1)
        else:
            info = self.client.request_account_tx(account,ledger_index_min,ledger_index_max)
        return info

    def account_lines(self, account):
        info = self.client.request_account_lines(account)
        return info

    def book_offers(self, taker_gets,taker_pays,limit = 10):
        TakerGets = {}
        TakerPays = {}
        if(taker_gets['currency'] == 'STM'):
            TakerGets = {"currency":"STM"}
        else:
            TakerGets = {"currency":taker_gets['currency'],"issuer":taker_gets['issuer']}
        if(taker_pays['currency'] == 'STM'):
            TakerPays = {"currency":"STM"}
        else:
            TakerPays = {"currency":taker_pays['currency'],"issuer":taker_pays['issuer']}
        info = self.client.book_offers(TakerGets,TakerPays,limit)
        return info

    def account_offers(self, account):
        info = self.client.account_offers(account)
        return info

    def tx(self, transaction):
        info = self.client.tx(transaction)
        return info

    def create_offer(self,taker_gets,taker_pays, account=None):
        if not account:
            if not self.secret:
                raise ValueError(
                    'If you do not provide a account, you need to '
                    'give a secret.')
            account = get_stream_from_secret(self.secret)
        tGets = Amount(taker_gets)
        tPays = Amount(taker_pays)
        if tGets.currency == 'STM':
            TakerGets = str(stm_drops(Decimal(tGets.value)))
        else:
            TakerGets = {
                "currency": tGets.currency,
                "issuer": tGets.issuer,
                "value": str(tGets.value)
            }
        if tPays.currency == 'STM':
            TakerPays = str(stm_drops(Decimal(tPays.value)))
        else:
            TakerPays = {
                "currency": tPays.currency,
                "issuer": tPays.issuer,
                "value": str(tPays.value)
            }
        tx = {
            "TransactionType" : "OfferCreate",
            "Account" : account,
            "TakerGets" : TakerGets,
            "TakerPays" : TakerPays
        }
        return self.submit(account, tx)

    def cancel_offer(self,sequence,account=None):
        if not account:
            if not self.secret:
                raise ValueError(
                    'If you do not provide a account, you need to '
                    'give a secret.')
            account = get_stream_from_secret(self.secret)
        tx = {
            "TransactionType" : "OfferCancel",
            "Account" : account,
            "OfferSequence" : Decimal(sequence)
        }
        return self.submit(account, tx)

    def send_payment(self, destination, amount, memo=None, destination_tag=None, account=None,
        flags=None):
        # Parse the amount
        amount = Amount(amount)

        # Determine the sender address
        if not account:
            if not self.secret:
                raise ValueError(
                    'If you do not provide a sender account, you need to '
                    'give a secret so one can be derived.')
            account = get_stream_from_secret(self.secret)

        # If the amount to send does not include an issuer, setting it
        # to the destination address makes stream pick one.
        if amount.currency != 'STM':
            if not 'issuer' in amount:
                amount['issuer'] = destination
        # For non-stm payments, we need to have a path
        '''
        if amount.currency == 'STM':
            paths = None
        else:
            pfr = self.client.find_path_once(
                source=account, destination=destination,
                destination_amount=amount
            )
            if not pfr['alternatives']:
                raise ValueError('No path found for this payment')
            paths = pfr['alternatives'][0]['paths_computed']
        '''

        tx = {
            "TransactionType" : "Payment",
            "Account" : account,
            "Destination" : destination,
            "Amount" : amount
        }

        if amount.currency != 'STM':
            sendMax = {}
            sendMax["currency"] = amount.currency
            sendMax["issuer"] = amount.issuer
            sendMax["value"] = str(Decimal(amount.value) * 12 / 10)
            tx["SendMax"] = sendMax
        else:
            tx["Amount"] = stm_drops(Decimal(amount.value))

        if destination_tag:
            tx['DestinationTag'] = destination_tag
        if flags is not None:
            tx['Flags'] = flags

        if memo is not None:
            if isinstance(memo,six.text_type) or isinstance(memo,six.binary_type):
                memoAdd = {}
                Memos = {}
                tx["Memos"] = []
                memoAdd["MemoType"] = '6D656D6F'
                if six.PY3:
                    memoData = six.binary_type(parse.quote(memo),'utf-8')
                else:
                    memoData = six.binary_type(parse.quote(memo))
                memoAdd["MemoData"] = six.text_type(fmt_hex(memoData))
                Memos["Memo"] = memoAdd
                tx["Memos"].append(Memos)
        #if paths is not None:
            #tx['Paths'] = paths

        return self.submit(account, tx)

    def submit(self, account, tx_json):
        """Returns a DeferredTransaction that you should wait() on.

        Or, raises an exception immediately during client.submit().
        """
        # Add a fee
        self.client.add_fee(tx_json)

        # Add sequence number
        tx_json['Sequence'] = self.get_sequence_number(account)

        # Sign the transaction
        sign_transaction(tx_json, self.secret)
        txhash = transaction_hash(tx_json)

        # Prepare a deferred result value
        pending = DeferredTransaction(tx_json, txhash)
        # Now submit
        result = self.client.submit(tx_blob=tx_json)

        # Let's deal with the result
        # This is analog to how stream-client deals with transactions.
        # Plus, see:
        error_code = result['engine_result']
        error_cat = error_code[:3]

        if error_cat == 'tec':
            # Fee was claimed, but transaction did not succeed.
            # Wiki claims this may be a proposed disposition, but JS client
            # just goes to error and finalizes.
            pending.resolve(result, error=error_code)
        elif error_cat == 'tes':
            # Success - proposed disposition.
            # JS client will emit an unused proposed event and then will
            # simply watch the transaction stream to confirm.
            self._pending_transactions[txhash] = pending
        elif error_cat == 'tef':
            # 'Failure': JS client will error out the transaction, unless
            # the message is tefPAST_SEQ: then it will resubmit three
            # ledgers later with a locally adjusted sequence number to
            # account for transactions happened in the intermediary.
            # TODO:  We don't do resubmission for now.
            pending.resolve(result, error=error_code)
        elif error_cat == 'ter':
            # Did not succeed initially, but may still, according to Wiki.
            # Despite this, the JS client first explicitly fetches a new
            # sequence number from the server and then resubmits. This
            # sounds like a potential race condition leading to a double
            # spend to me.
            # TODO: Anyway, we don't do resubmission for now.
            pending.resolve(result, error=error_code)
        else:
            # Default - must be tem (malFormed) or tel (local); those are
            # final failures.
            # TODO: JS client resubmits on tooBusy one ledger later
            pending.resolve(result, error=error_code)
            self._sequence_cache[account] -= 1

        return pending



from __future__ import print_function
import six
if six.PY2:
    import urllib as parse
else:
    import urllib.parse as parse
import os
import json
import time
import datetime
from decimal import Decimal
from .sign import get_secret,get_stream_from_secret
from .client import Remote, ResponseError,transaction_hash
from .datastructures import Amount,stm,stm_drops

LOCAL_SIGNING = int(os.environ.get('LOCAL_SIGNING', 1))
TESTNET = int(os.environ.get('testnet', 0))
WSS_URL= 'wss://node.labs.stream'
if TESTNET == 1:
    WSS_URL = 'wss://sa.labs.stream'

def generate():
    '''
    generate new address and secret pair
    '''
    secret = get_secret()
    output = {
        'address': get_stream_from_secret(secret),
        'secret': secret
    }
    return output

def send(secret, destination, amount, memo=None, dt=None):
    '''
    send in stream network
    '''
    remote = Remote(WSS_URL,secret)
    result = remote.send_payment(destination,amount,memo,dt)
    try:
        result = result.wait()
    except ResponseError as e:
        result = e.response
    if True:
        res = {}
        if 'transaction' in result:
            tx = result['transaction']
        else:
            tx = result['tx_json']
        res['from'] = tx['Account']
        res['to'] = tx['Destination']
        res['amount'] = {}
        amount = Amount(tx['Amount'])
        res['amount']['currency'] = amount.currency
        res['amount']['value'] = str(amount.value)
        res['amount']['issuer'] = amount.issuer
        if dt:
            res['dt'] = dt
        if memo:
            res['memo'] = memo
        res['tx'] = {}
        res['tx']['engine_result'] = result['engine_result']
        res['tx']['engine_result_message'] = result['engine_result_message']
        res['tx']['hash'] = tx['hash']
        return res

def toLocaltime(rpepoch):
    '''
    convert strem time to localtime
    '''
    timeStamp =  rpepoch + 0x386D4380
    value = time.localtime(timeStamp)
    formatedTime = time.strftime('%Y-%m-%d %H:%M:%S',value)
    return formatedTime

def parseAmount(amount):
    result = {}
    if isinstance(amount,dict):
        result['currency'] = amount['currency']
        result['issuer'] = amount['issuer']
        result['value'] = amount['value']
    elif isinstance(amount,(six.binary_type,six.text_type,int)):
        result['currency'] = 'STM'
        result['issuer'] = ''
        result['value'] = str(stm(amount))
    return result

def subtractAmount(amount1,amount2):
    result = {}
    if(amount1['currency'] == amount2['currency'] and amount1['issuer'] == amount2['issuer']):
        result['currency'] = amount1['currency']
        result['issuer'] = amount1['issuer']
        result['value'] = str(float(amount1['value'])-float(amount2['value']))
    return result

def processNodes(AffectedNodes):
    effects = []
    for nodes in AffectedNodes:
        effect = {}
        for key in nodes:
            if(key  == 'CreatedNode' or key == 'ModifiedNode' or key == 'DeletedNode'):
                effect['diffType'] = key
                effect['entryType'] = nodes[key]['LedgerEntryType']
                effect['ledgerIndex'] = nodes[key]['LedgerIndex']
                effect['fields'] = {}
                if 'PreviousFields' in  nodes[key]:
                    effect['PreviousFields'] = nodes[key]['PreviousFields']
                    effect['fields'].update(nodes[key]['PreviousFields'])
                if 'FinalFields' in nodes[key]:
                    effect['FinalFields'] = nodes[key]['FinalFields']
                    effect['fields'].update(nodes[key]['FinalFields'])
                if 'NewFields' in nodes[key]:
                    effect['NewFields'] = nodes[key]['NewFields']
                    effect['fields'].update(nodes[key]['NewFields'])
                effects.append(effect)
    return effects

def processTxn(tx, meta, address):
    '''
    parse the transaction detail
    '''
    obj = {}
    obj['effects'] = []

    ## Main transaction
    if (tx['Account'] == address or ('Destination' in tx and tx['Destination'] == address)):
        if 'tesSUCCESS' == meta['TransactionResult']:
            if tx['TransactionType'] == 'Payment':
                obj['account'] = address
                if tx['Account'] == address:
                    if tx['Destination'] == address:
                        obj['type'] = 'exchange'
                    else:
                        obj['type'] = 'sent'
                        obj['counterparty'] = tx['Destination']
                        counterparty = tx['Destination']
                else:
                    obj['type'] = 'received'
                    obj['counterparty'] = tx['Account']
                    counterparty = tx['Account']
                delivered_amount = meta['delivered_amount']
                if (type(delivered_amount) == type({}) and 'currency' in delivered_amount):
                    obj['amount'] = meta['delivered_amount']['value']
                    obj['currency'] = meta['delivered_amount']['currency']
                else:
                    obj['amount'] = str(Decimal(meta['delivered_amount'])/1000000)
                    obj['currency'] = 'STM'

                for affectedNode in meta['AffectedNodes']:
                    if 'ModifiedNode' in affectedNode:
                        ModifiedNode = affectedNode['ModifiedNode']
                        if (type(ModifiedNode) == type({}) and ModifiedNode['LedgerEntryType'] == 'VStreamState' and (ModifiedNode['FinalFields']['HighLimit']['issuer'] == counterparty or ModifiedNode['FinalFields']['LowLimit']['issuer'] == counterparty) and ModifiedNode['FinalFields']['Balance']['currency'] == obj['currency']):
                            if ModifiedNode['FinalFields']['HighLimit']['issuer']==counterparty:
                                obj['issuer'] = ModifiedNode['FinalFields']['LowLimit']['issuer']
                            else:
                                obj['issuer'] = ModifiedNode['FinalFields']['HighLimit']['issuer']

                if 'DestinationTag' in tx:
                    obj['DestinationTag'] = str(tx['DestinationTag'])

                if ('Memos' in tx and tx['Memos'][0]) :
                    data = str(bytearray.fromhex(tx['Memos'][0]['Memo']['MemoData']).decode())
                    obj['memoData'] = parse.unquote(data)
            if tx['TransactionType'] == 'OfferCreate':
                obj['type'] = 'offernew'
                obj['offer'] = {}
                obj['offer']['sequence'] = tx['Sequence']

                obj['offer']['Flags'] = tx['Flags']
                obj['offer']['buy'] = parseAmount(tx['TakerPays'])
                obj['offer']['sell'] = parseAmount(tx['TakerGets'])
            if tx['TransactionType'] == 'OfferCancel':
                obj['type'] = 'offercancel'

    if 'tesSUCCESS' == meta['TransactionResult']:

        nodes = processNodes(meta['AffectedNodes'])
        obj['funded'] = 'none'
        for node in nodes:
            effect = {}
            effect['offer'] = {}
            if node['entryType'] == 'Offer':
                fieldSet = node['fields']
                if node['fields']['Account'] == address:
                    if(node['diffType'] == 'ModifiedNode' or (node['diffType'] == 'DeletedNode' and 'PreviousFields' in node and 'TakerGets' in node['PreviousFields'] and parseAmount(node['FinalFields']['TakerGets']))):
                        effect['type'] = 'offer_partially_funded'
                        if node['diffType'] == 'DeletedNode':
                            obj['funded'] = 'full'
                    else:
                        if node['diffType'] == 'CreatedNode':
                            effect['type'] = 'offer_created'
                        elif('PreviousFields' in node and 'TakerPays' in node['PreviousFields']):
                            effect['type'] = 'offer_funded'
                        else:
                            effect['type'] = 'offer_cancelled'
                        if effect['type'] == 'offer_funded':
                            fieldSet = node['PreviousFields']
                        if(effect['type'] == 'offer_cancelled' and obj['type'] == 'offercancel'):
                            if 'offer' not in obj:
                                obj['offer'] = {}
                            obj['offer']['sell'] = parseAmount(fieldSet['TakerGets'])
                            obj['offer']['buy'] = parseAmount(fieldSet['TakerPays'])

                elif(tx['Account'] == address and 'PreviousFields' in node and node['PreviousFields']):
                    effect['type'] = 'offer_bought'
                if 'type' in effect:
                    if 'remaining' not in effect:
                        effect['remaining'] = {}
                    effect['remaining']['sell'] = parseAmount(fieldSet['TakerGets'])
                    effect['remaining']['buy'] = parseAmount(fieldSet['TakerPays'])
                    if effect['remaining']['sell']['value'] =='0' and effect['remaining']['buy']['value'] == '0':
                        del effect['remaining']
                    if(effect['type'] == 'offer_partially_funded' or effect['type'] == 'offer_bought'):
                        if 'offer' not in obj:
                            obj['offer'] = {}
                        obj['offer']['sequence'] = fieldSet['Sequence']
                        if 'fulfilled' not in effect:
                            effect['fulfilled'] = {}
                        obj['funded'] = 'partial'
                        effect['fulfilled']['got'] = subtractAmount(parseAmount(node['PreviousFields']['TakerGets']),parseAmount(node['fields']['TakerGets']))
                        effect['fulfilled']['paid'] = subtractAmount(parseAmount(node['PreviousFields']['TakerPays']),parseAmount(node['fields']['TakerPays']))
                if 'offer' in obj:
                    if 'PreviousTxnID' in fieldSet:
                        obj['offer']['hash'] = fieldSet['PreviousTxnID']
                    elif 'buy' in obj['offer']:
                        obj['offer']['hash'] = str(tx['hash'])
                    if 'sequence' not in obj['offer']:
                        obj['offer']['sequence'] = fieldSet['Sequence']
            if 'offer' in effect and not effect['offer']:
                del effect['offer']
            if 'type' in effect:
                obj['effects'].append(effect)

    obj['hash'] = str(tx['hash'])
    obj['date'] = tx['date']
    obj['time'] = toLocaltime(tx['date'])
    obj['ledger_index'] = tx['ledger_index']
    obj['tx_type'] = tx['TransactionType']

    if ('effects' in obj and len(obj['effects']) == 0):
        del obj['effects']
    return obj

def getPaymentHistory(address,ledger_index_min = -1,ledger_index_max = -1):
    '''
    get history from stream network
    '''
    remote = Remote(WSS_URL,'')
    result = remote.account_tx(address,ledger_index_min,ledger_index_max)
    transactions = []
    for transaction in result['transactions']:
        tx = processTxn(transaction['tx'],transaction['meta'],address)
        if tx and tx['tx_type'] == 'Payment':
            transactions.append(tx)
    res = {}
    res['account'] = address
    res['ledger_index_max'] = result['ledger_index_max']
    res['transactions'] = transactions
    return res

def getOfferHistory(address,ledger_index_min = -1,ledger_index_max = -1):
    '''
    get history from stream network
    '''
    remote = Remote(WSS_URL,'')
    result = remote.account_tx(address,ledger_index_min,ledger_index_max)
    transactions = []
    for transaction in result['transactions']:
        tx = processTxn(transaction['tx'],transaction['meta'],address)
        if tx and (tx['tx_type'] == 'OfferCancel' or tx['tx_type'] == 'OfferCreate'):
            transactions.append(tx)
    res = {}
    res['account'] = address
    res['ledger_index_max'] = result['ledger_index_max']
    res['transactions'] = transactions
    return res

def getLines(address):
    '''
    get trustlines and stm balance from stream network
    '''
    remote = Remote(WSS_URL,'')
    lines = remote.account_lines(address)
    stmline = remote.account_info(address)
    res = {}
    res['account'] = address
    res['balance'] = str(stm(stmline['Balance']))
    res['trustLines'] = []
    for line in lines:
        oneLine = {}
        oneLine['issuer'] = line['account']
        oneLine['balance'] = line['balance']
        oneLine['currency'] = line['currency']
        oneLine['limit'] = line['limit']
        res['trustLines'].append(oneLine)
    return res

def book_offers(sell,buy):
    '''
    get 10 offers from stream network
    '''
    remote = Remote(WSS_URL,'')
    offerResult = remote.book_offers(sell,buy,10)
    offers = []
    for offerLine in offerResult['offers']:
        offer = {}
        offer['sell'] = parseAmount(offerLine['TakerGets'])
        offer['buy'] = parseAmount(offerLine['TakerPays'])
        if offer['sell']['currency'] == 'STM':
            offer['price'] = str(stm_drops(offerLine['quality']))
        elif offer['buy']['currency'] == 'STM':
            offer['price'] = str(stm(offerLine['quality']))
        else:
            offer['price'] = str(Decimal(offerLine['quality']))
        if offer['sell']['value'] != '0' and offer['buy']['value'] != '0':
            offers.append(offer)
    return offers

def market_offers(pair_counterparty,pair_base):
    '''
    get offers from stream newwork
    '''
    buy = book_offers(pair_base,pair_counterparty)
    sell = book_offers(pair_counterparty,pair_base)
    result = {}
    result['pair'] = pair_counterparty['currency']+'/'+pair_counterparty['issuer']+'_'+pair_base['currency']+'/'+pair_base['issuer']
    if len(buy) > 0:
        result['buy'] = []
        for offer in buy:
            buyoffer = {}
            buyoffer['amount'] = offer['buy']['value']
            buyoffer['price'] = str(Decimal(offer['sell']['value'])/Decimal(offer['buy']['value']))
            result['buy'].append(buyoffer)
    if len(sell) > 0:
        result['sell'] = []
        for offer in sell:
            selloffer = {}
            selloffer['amount'] = offer['sell']['value']
            selloffer['price'] = str(Decimal(offer['buy']['value'])/Decimal(offer['sell']['value']))
            result['sell'].append(selloffer)
    return result

def account_offers(address):
    '''
    get offers of the account from stream network
    '''
    remote = Remote(WSS_URL,'')
    offerResult = remote.account_offers(address)
    offers = []
    for offerLine in offerResult['offers']:
        offer = {}
        offer['sequence'] = offerLine['seq']
        offer['sell'] = parseAmount(offerLine['taker_gets'])
        offer['buy'] = parseAmount(offerLine['taker_pays'])
        if offer['buy']['value'] != '0' or offer['sell']['value'] != '0':
            offers.append(offer)
    return offers

def create_offer(secret,sell,buy):
    '''
    create a offer
    '''
    remote = Remote(WSS_URL,secret)
    result = remote.create_offer(sell,buy)
    try:
        result = result.wait()
    except ResponseError as e:
        result = e.response

    if True:
        res = {}
        res['tx'] = {}
        res['tx']['offer'] = {}
        res['tx']['offer']['TakerGets'] = sell
        res['tx']['offer']['TakerPays'] = buy
        res['tx']['engine_result'] = result['engine_result']
        res['tx']['engine_result_message'] = result['engine_result_message']
        if result['engine_result'] == 'tesSUCCESS':
            res['tx']['hash'] = result['transaction']['hash']
            res['tx']['sequence'] = result['transaction']['Sequence']
        return res

def cancel_offer(secret, sequence):
    '''
    cancel a offer with the sequence in stream network
    '''
    remote = Remote(WSS_URL,secret)
    result = remote.cancel_offer(sequence)
    try:
        result = result.wait()
    except ResponseError as e:
        result = e.response

    if True:
        res = {}
        res['tx'] = {}
        res['tx']['engine_result'] = result['engine_result']
        res['tx']['engine_result_message'] = result['engine_result_message']
        if result['engine_result'] == 'tesSUCCESS':
            res['tx']['sequence'] = sequence
            res['tx']['hash'] = result['transaction']['hash']
        return res

def tx(address,transaction):
    '''
    get the detail of a tx
    '''
    remote = Remote(WSS_URL,'')
    detail = remote.tx(transaction)
    return processTxn(detail,detail['meta'],address)



#!/usr/bin/env python
# coding: utf-8
import sys
from setuptools import setup, find_packages


install_requires = [
    'ecdsa>=0.10',
    'six>=1.5.2',
    'websocket-client==0.15.0'
]


setup(
    name="stmpy_lib",
    version="0.1.0b",
    author="stm-open",
    author_email="open@labs.stream",
    url="https://github.com/stm-open/stmpy.git",
    description="Python lib for the STM network",
    license='BSD',
    packages=find_packages(),
    zip_safe=True,
    install_requires=install_requires,
)
